import json
import os
import random
import secrets
import string
from dataclasses import dataclass
from enum import Enum
from typing import Optional

import boto3
from mypy_boto3_ssm import SSMClient
from sqlalchemy.future.engine import Connection
from sqlmodel import text


class Privilege(str, Enum):
    """
    Possible privileges to grant to Postgres users: https://www.postgresql.org/docs/15/ddl-priv.html
    """

    SELECT = "SELECT"
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    TRUNCATE = "TRUNCATE"
    REFERENCES = "REFERENCES"
    TRIGGER = "TRIGGER"
    CREATE = "CREATE"
    CONNECT = "CONNECT"
    TEMPORARY = "TEMPORARY"
    EXECUTE = "EXECUTE"
    USAGE = "USAGE"
    SET = "SET"
    ALTER_SYSTEM = "ALTER_SYSTEM"


class Stage(str, Enum):
    dev = "dev"
    prod = "prod"


class Partner(str, Enum):
    afro_la = "afro_la"
    dallas_free_press = "dallas_free_press"
    open_vallejo = "open_vallejo"
    the_19th = "the_19th"


@dataclass
class Grant:
    privileges: list[Privilege]
    tables: list[str]


@dataclass
class RowLevelSecurityPolicy:
    # TODO have table and user_column reference actual Table and Column objects
    table: str
    user_column: str
    policy_name: Optional[str] = None


@dataclass
class Component:
    # effectively a db-constrained role
    name: str
    grants: list[Grant]
    policies: list[RowLevelSecurityPolicy]


@dataclass
class DatabaseConnectionCredentials:
    HOST: str
    PORT: int
    DB_NAME: str
    USERNAME: str
    PASSWORD: str


def create_database(conn: Connection, db_name: Stage) -> None:
    statement = text(f"CREATE DATABASE {db_name}")
    conn.execution_options(isolation_level="AUTOCOMMIT").execute(statement)


def create_role(conn: Connection, role: str) -> None:
    statement = text(f"CREATE ROLE {role}")
    conn.execute(statement)


def generate_password() -> str:
    alphabet = string.ascii_letters + string.digits
    while True:
        password = "".join(secrets.choice(alphabet) for i in range(10))
        if (
            any(c.islower() for c in password)
            and any(c.isupper() for c in password)
            and sum(c.isdigit() for c in password) >= 3
        ):
            break
    return password


def create_user(conn: Connection, username: str) -> str:
    pw = generate_password()
    statement = text(f"CREATE USER {username} WITH PASSWORD :password")
    conn.execute(statement, {"password": pw})
    return pw


def assign_role(conn: Connection, role: str, usernames: list[str]) -> None:
    users = ", ".join(usernames)
    statement = text(f"GRANT {role} TO {users}")
    conn.execute(statement)


def grant_privileges(conn: Connection, user_or_role: str, table: str, privileges: list[Privilege]) -> None:
    formatted_privileges = ", ".join(privileges)
    statement = text(f"GRANT {formatted_privileges} ON {table} TO {user_or_role}")
    conn.execute(statement)


def enable_row_level_security(conn: Connection, table: str, target_column: str, role: str) -> None:
    s1 = text(f"ALTER TABLE {table} ENABLE ROW LEVEL SECURITY")
    conn.execute(s1)
    random_postfix = "%06x" % random.randrange(16**6)
    policy_name = f"{table}_{role}_{random_postfix}"
    s2 = text(f"CREATE POLICY {policy_name} ON {table} TO {role} USING (current_user LIKE '%' || {target_column} || '%')")
    conn.execute(s2)


def create_users(conn: Connection, stage: Stage, component: Component, partner_names: list[Partner]) -> None:
    ssm_client = get_ssm_client()
    for partner_name in partner_names:
        username = f"{stage}_{component.name}_{partner_name}"
        pw = create_user(conn, username=username)
        if ssm_client:
            user_creds = DatabaseConnectionCredentials(
                HOST=os.getenv("HOST", "localhost"),
                PORT=int(os.getenv("PORT", "5432")),
                DB_NAME=stage,
                USERNAME=username,
                PASSWORD=pw,
            )
            ssm_client.put_parameter(
                Name=f"/{stage}/ata/{partner_name}/{component.name}/database-credentials",
                Description=f"DB credentials for partner {partner_name}, component {component}, and env {stage}.",
                Value=json.dumps(user_creds.__dict__),
                Type="String",
                Overwrite=True,
            )


def get_conn_string(db_name: Optional[str] = None) -> str:
    # everything but dbname should be the same, since we are using the admin user for everything
    host = os.getenv("HOST", "localhost")
    port = os.getenv("PORT", "5432")
    user = os.getenv("USERNAME", "postgres")
    password = os.getenv("PASSWORD", "postgres")
    if not db_name:
        # if no db_name is passed, we assume it is for the default db. This is assumed to be "default"
        # unless indicated otherwise via this DB_NAME env var
        db_name = os.getenv("DB_NAME", "default")

    return f"postgresql://{user}:{password}@{host}:{port}/{db_name}"


def get_ssm_client() -> Optional[SSMClient]:
    # ENABLE_SSM has to explicitly be set to exactly "TRUE" or else no SSM interactions take place
    if os.getenv("ENABLE_SSM", "FALSE") == "TRUE":
        return boto3.client("ssm")
    else:
        return None



from sqlalchemy.future.engine import create_engine

from ata_db_models.helpers import (
    Component,
    Grant,
    Partner,
    Privilege,
    RowLevelSecurityPolicy,
    Stage,
    assign_role,
    create_database,
    create_role,
    create_users,
    enable_row_level_security,
    get_conn_string,
    grant_privileges,
)
from ata_db_models.models import SQLModel


def pre_table_initialization(stage: Stage, components: list[Component], partner_names: list[Partner]) -> None:
    engine = create_engine(get_conn_string())

    with engine.connect() as conn:
        create_database(conn, db_name=stage)
        for component in components:
            role = f"{stage}_{component.name}"
            create_role(conn, role=role)
            usernames = [f"{stage}_{component.name}_{partner_name}" for partner_name in partner_names]
            create_users(conn, stage=stage, component=component, partner_names=partner_names)
            assign_role(conn, role=role, usernames=usernames)
        conn.commit()


def initialize_tables(stage: Stage) -> None:
    engine = create_engine(get_conn_string(db_name=stage))
    SQLModel.metadata.create_all(engine)


def post_table_initialization(stage: Stage, components: list[Component]) -> None:
    engine = create_engine(get_conn_string(db_name=stage))

    with engine.connect() as conn:
        for component in components:
            for grant in component.grants:
                for table in grant.tables:
                    grant_privileges(
                        conn, user_or_role=f"{stage}_{component.name}", table=table, privileges=grant.privileges
                    )
            for policy in component.policies:
                enable_row_level_security(
                    conn, table=policy.table, target_column=policy.user_column, role=f"{stage}_{component.name}"
                )

        conn.commit()


def initialize_all_database_entities(stage: Stage, components: list[Component], partner_names: list[Partner]) -> None:
    # needs default db
    pre_table_initialization(stage=stage, components=components, partner_names=partner_names)
    # needs to connect to this stage's db
    initialize_tables(stage=stage)
    # needs this stage's db
    post_table_initialization(stage=stage, components=components)


if __name__ == "__main__":
    stages = [Stage.dev, Stage.prod]
    pipeline0 = Component(
        name="pipeline0",
        grants=[
            Grant(privileges=[Privilege.SELECT, Privilege.INSERT, Privilege.UPDATE, Privilege.DELETE], tables=["event"])
        ],
        policies=[RowLevelSecurityPolicy(table="event", user_column="site_name")],
    )
    components = [pipeline0]
    partner_names = [Partner.afro_la, Partner.dallas_free_press, Partner.open_vallejo, Partner.the_19th]

    # roles are dev-pipeline0 and prod-pipeline0
    # users are dev-pipeline0-afro-la, dev-pipeline0-dallas-free-press, ... , prod-pipeline0-the-19th

    for stage in stages:
        initialize_all_database_entities(stage=stage, components=components, partner_names=partner_names)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ata_db_models']

package_data = \
{'': ['*']}

install_requires = \
['alembic==1.8.1',
 'boto3-stubs[ssm]==1.26.13',
 'boto3==1.26.13',
 'psycopg2==2.9.5',
 'sqlmodel==0.0.8']

setup_kwargs = {
    'name': 'ata-db-models',
    'version': '0.0.10',
    'description': 'Database models and migrations for Automating the Ask.',
    'long_description': "# ata-db-models\n\n<!-- [![Release](https://img.shields.io/github/v/release/LocalAtBrown/ata-db-models)](https://img.shields.io/github/v/release/LocalAtBrown/ata-db-models) -->\n<!-- [![Build status](https://img.shields.io/github/workflow/status/LocalAtBrown/ata-db-models/merge-to-main)](https://img.shields.io/github/workflow/status/LocalAtBrown/ata-db-models/merge-to-main) -->\n\n[![Python version](https://img.shields.io/badge/python_version-3.9-blue)](https://github.com/psf/black)\n[![Code style with black](https://img.shields.io/badge/code_style-black-000000.svg)](https://github.com/psf/black)\n[![More style with flake8](https://img.shields.io/badge/code_style-flake8-blue)](https://flake8.pycqa.org)\n[![Imports with isort](https://img.shields.io/badge/%20imports-isort-blue)](https://pycqa.github.io/isort/)\n[![Type checking with mypy](https://img.shields.io/badge/type_checker-mypy-blue)](https://mypy.readthedocs.io)\n[![License](https://img.shields.io/github/license/LocalAtBrown/ata-db-models)](https://img.shields.io/github/license/LocalAtBrown/ata-db-models)\n\nDatabase models and migrations for Automating the Ask.\n\n## Usage\n\n### As a package\n\nWe use [SQLModel](https://sqlmodel.tiangolo.com/), a layer on top of SQLAlchemy with Pydantic, to define our tables.\nThis is useful because we can import this package to interact with the tables AND have Pydantic objects in Python\nthat correspond to a row in the table.\n\nTo install the package from PyPi, run: `pip install ata-db-models`. Check existing versions \n[here](https://pypi.org/project/ata-db-models/).\n\n### Initialize a new cluster\n\nIf you want to initialize a fresh database cluster, pass in the env vars to connect to the cluster and run `init_db`.\nIf the target cluster has IP restrictions, make sure your IP address is a valid access point.\n\nAn example run with fake credentials (from the root dir of this project with the virtual env\nactivated):\n`HOST=fakehost USER=fakeuser PASSWORD=fakepw DB_NAME=postgres python src/init_db.py`\n\nNo `PORT` is passed because the default port is 5432, the standard for Postgres.\n\n### Migrations\n\nSo you made some changes to what tables there are, what columns there are, indices, etc. and you'd like to\nupdate the databases. This is what alembic is for!\n\n TODO ALEMBIC INSTRUCTIONS\n\n## Development\n\nThis project uses [Poetry](https://python-poetry.org/) to manage dependencies. It also helps with pinning dependency and python\nversions. We also use [pre-commit](https://pre-commit.com/) with hooks for [isort](https://pycqa.github.io/isort/),\n[black](https://github.com/psf/black), and [flake8](https://flake8.pycqa.org/en/latest/) for consistent code style and\nreadability. Note that this means code that doesn't meet the rules will fail to commit until it is fixed.\n\nWe use [mypy](https://mypy.readthedocs.io/en/stable/index.html) for static type checking. This can be run [manually](#run-static-type-checking),\nand the CI runs it on PRs to the `main` branch. We also use [pytest](https://docs.pytest.org/en/7.2.x/) to run our tests.\nThis can be run [manually](#run-tests) and the CI runs it on PRs to the `main` branch.\n\n### Setup\n\n1. [Install Poetry](https://python-poetry.org/docs/#installation).\n2. Run `poetry install --no-root`\n3. Run `source $(poetry env list --full-path)/bin/activate && pre-commit install && deactivate` to set up `pre-commit`\n\nYou're all set up! Your local environment should include all dependencies, including dev dependencies like `black`.\nThis is done with Poetry via the `poetry.lock` file.\n\n### Run Code Format and Linting\n\nTo manually run isort, black, and flake8 all in one go, simply run `pre-commit run --all-files`. Explore the `pre-commit` docs (linked above)\nto see more options.\n\n### Run Static Type Checking\n\nTo manually run mypy, simply run `mypy` from the root directory of the project. It will use the default configuration\nspecified in `pyproject.toml`.\n\n### Update Dependencies\n\nTo update dependencies in your local environment, make changes to the `pyproject.toml` file then run `poetry update` from the root directory of the project.\n\n### Run Tests\n\nTo manually run rests, you need to have a Postgres instance running locally on port 5432. One way to do this\nis to run a Docker container, then run the tests while it is active.\n\n1. (If you don't already have the image locally) Run `docker pull postgres`\n2. Run `docker run --rm --name postgres -e POSTGRES_PASSWORD=postgres -e POSTGRES_HOST_AUTH_METHOD=trust -p 127.0.0.1:5432:5432/tcp postgres`\n3. Run `DB_NAME=postgres pytest tests` from the root directory of the project. Explore the `pytest` docs (linked above)\nto see more options.\n\nNote that if you decide to run the Postgres container with different credentials (a different password, port, etc.) or\nvia a different method, you will likely need to update the test file to point to the correct Postgres instance.\n\nAdditionally, if you want to re-run the tests, you want to make sure you start over from a fresh Postgres\ninstance. If you run Postgres via Docker, you can simply `ctrl-C` to stop the image and start a new one.\n",
    'author': 'Raaid Arshad',
    'author_email': 'raaid@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/LocalAtBrown/ata-models',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)
